import hashlib
import json
import shutil
from pathlib import Path
from typing import Dict, Optional

from common_utils.core.base import Storage


# pylint: disable=line-too-long
class SimpleDVC:
    """
    # TODO: try to see if we can make it work for local storage as well.

    Local Scenario 1:
    pipeline-training/data
    ├── interim
    ├── processed
    └── raw
        ├── filtered_movies_incremental.csv             # filename
        └── filtered_movies_incremental.csv.json        # metadata for filename

    This scenario handles the workflow whereby user will dump the below stores folder
    into remote storage during runtime. For example, if user chooses not to store
    the actual data and just the .csv.json metadata file, then the below will be dumped
    into say, a GCS bucket or MLFlow artifact store.
    Of course, ideally, the metadata store in the same location as the actual data
    so git can track it.
    Local Scenario 2:
    /pipeline-training/stores/<UNIQUE_RUN_ID>
    |── logs
    |── registry
    |── artifacts
    └── blobs
        ├── raw
        |   ├── filtered_movies_incremental.csv.json         # metadata for filename


    So in this case:
    - data_dir = /pipeline-training/data
    - metadata_dir = /pipeline-training/stores/<UNIQUE_RUN_ID>/blobs/raw

    REMOTE DVC TREE
    gaohn/                                              # remote_bucket_name
    └── imdb                                            # remote_bucket_project_name
        └── gaohn-dvc                                   # remote_dvc_dir_name
            ├── 1234567890                              # md5 hash for raw/filtered_movies_incremental.csv
            └── filtered_movies_incremental.csv.json
    """

    def __init__(
        self,
        storage: Storage,
        remote_bucket_project_name: str,
        data_dir: str = "./data",
        cache_dir: str = ".cache",
        metadata_dir: Optional[
            str
        ] = None,  # new parameter for the location of metadata
    ) -> None:
        self.storage = storage
        self._remote_bucket_project_name = remote_bucket_project_name

        self.data_dir = Path(data_dir)

        self.cache_dir = Path(cache_dir)
        self.cache_dir.mkdir(parents=True, exist_ok=True)

        # if metadata_dir is None, then metadata_dir = data_dir
        self.metadata_dir = Path(metadata_dir) if metadata_dir else self.data_dir

        self.metadata_file: Path

    @property
    def remote_dvc_dir_name(self) -> str:
        """Always fixed to gaohn-dvc. Immutable just like how dvc always uses .dvc."""
        return "gaohn-dvc"

    @property
    def remote_bucket_name(self) -> str:
        """It is important to know that the Storage object has `bucket_name` as an attribute."""
        return self.storage.bucket_name

    @property
    def remote_bucket_project_name(self) -> str:
        return self._remote_bucket_project_name

    def _get_destination_blob_name(self, md5: str) -> str:
        return f"{self.remote_bucket_project_name}/{self.remote_dvc_dir_name}/{md5}"

    def _create_gitignore(self, pattern: str) -> None:
        gitignore_file = self.data_dir / ".gitignore"
        if not gitignore_file.exists():
            with open(gitignore_file, "w", encoding="utf-8") as file:
                file.write(pattern)
        else:
            with open(gitignore_file, "a+", encoding="utf-8") as file:
                file.seek(0)
                if pattern not in file.read():
                    file.write("\n" + pattern)

    def _get_cache_file_path(self, filename: str) -> Path:
        return self.cache_dir / filename

    def _get_metadata_file_path(self, filename: str) -> Path:
        """
        Constructs a Path object for a metadata file in the data directory.

        Args:
            filename (str): Name of the file.

        Returns:
            Path: Path object for the metadata file in the data directory.
        """
        return self.metadata_dir / f"{filename}.json"

    def _load_metadata(self, filename: str) -> Dict[str, str]:
        """
        Loads the metadata for a file from its metadata JSON file.

        Args:
            filename (str): Name of the file.

        Returns:
            Dict[str, str]: Metadata dictionary for the file.
        """
        metadata_file = self._get_metadata_file_path(filename)
        with metadata_file.open("r") as file:
            return json.load(file)

    def _calculate_md5(self, filepath: str) -> str:
        hash_md5 = hashlib.md5()
        with open(filepath, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def add(self, filepath: str) -> Dict[str, str]:
        filepath = Path(filepath)
        filename = filepath.name
        extension = filepath.suffix

        md5 = self._calculate_md5(str(filepath))
        cache_filepath = self.cache_dir / md5

        shutil.copy(filepath, cache_filepath)

        metadata = {
            "filename": filename,
            "filepath": str(cache_filepath),
            "data_dir": str(self.data_dir),
            "metadata_dir": str(self.metadata_dir),
            "extension": extension,
            "remote_bucket_name": self.remote_bucket_name,
            "remote_bucket_project_name": self.remote_bucket_project_name,
            "remote_dvc_dir_name": self.remote_dvc_dir_name,
            "md5": md5,
        }

        self.metadata_file = self._get_metadata_file_path(filename)
        with self.metadata_file.open("w") as file:
            json.dump(metadata, file, indent=4)

        self._create_gitignore(filename)

        return metadata

    def push(self, filepath: str) -> None:
        filepath = Path(filepath)
        filename = filepath.name
        metadata = self._load_metadata(filename)
        cache_filepath = metadata["filepath"]

        md5 = metadata["md5"]

        destination_blob_name = self._get_destination_blob_name(md5)

        # This effectively uploads the cached file to the remote storage
        self.storage.upload_blob(
            source_file_name=cache_filepath, destination_blob_name=destination_blob_name
        )

    def pull(self, filename: str, remote_bucket_project_name: str) -> None:
        metadata = self._load_metadata(filename)
        source_blob_name = (
            f"{remote_bucket_project_name}/{self.remote_dvc_dir_name}/{metadata['md5']}"
        )
        self.storage.download_blob(source_blob_name, self.data_dir / filename)

    def checkout(self, filename: str) -> None:
        metadata = self._load_metadata(filename)
        if metadata["filename"] == filename:
            shutil.copy(metadata["filepath"], str(self.data_dir / filename))



"""Core functions for Git versioning.

TODO:
    - see if it setting working_dir = None is the same as working_dir = ".".
        if so, use working_dir = working_dir or ".".
"""
import subprocess
from typing import List, Literal, Optional, Union

from common_utils.core.logger import Logger

# Setup logging
logger = Logger(
    module_name=__name__, propagate=False, log_root_dir=None, log_file=None
).logger


def check_git_status(working_dir: Optional[str] = None) -> Literal[True, False]:
    """
    Check the Git status of the working directory. If there are untracked or
    uncommitted changes, return False.

    Parameters
    ----------
    working_dir : str, optional
        The path of the working directory where the Git command should be executed,
        by default None. If None, it uses the current working directory.

    Returns
    -------
    Literal[True, False]
        True if there are no untracked or uncommitted changes in the working directory,
        False otherwise.
    """
    status_output = (
        subprocess.check_output(["git", "status", "--porcelain"], cwd=working_dir)
        .decode("utf-8")
        .strip()
    )
    return len(status_output) == 0


def log_message_if_working_dir_is_none(working_dir: Optional[str] = None) -> None:
    """
    Log an informative message if no working directory is provided.

    Parameters
    ----------
    working_dir : str, optional
        The path of the working directory, by default None.
            If None, an info message is logged.
    """
    if working_dir is None:
        logger.info("Working directory not provided. Defaulting to current directory.")
    logger.info(f"Working directory: {working_dir}")


def get_git_commit_hash(
    working_dir: Optional[str] = None, check_git_status: Literal[True, False] = False
) -> Union[str, Literal["N/A"]]:
    """
    Get the current Git commit hash.

    If Git is not installed or the working directory is not a Git repository,
    the function returns "N/A".

    Parameters
    ----------
    working_dir : str, optional
        The path of the working directory where the Git command should be executed,
        by default None. If None, it uses the current working directory ".".
    check_git_status : Literal[True, False], optional
        Whether to check if there are untracked or uncommitted changes in the
        working directory, by default False.

    Returns
    -------
    commit_hash : str
        The Git commit hash, or "N/A" if Git is not installed or the working
        directory is not a Git repository.
    """
    log_message_if_working_dir_is_none(working_dir)

    git_command = ["git", "rev-parse", "HEAD"]

    try:
        if check_git_status and not check_git_status(working_dir):
            error_message = (
                "There are untracked or uncommitted files in the working directory. "
                "Please commit or stash them before running training as the commit hash "
                "will be used to tag the model."
            )
            raise RuntimeError(error_message)

        commit_hash = (
            subprocess.check_output(git_command, cwd=working_dir)
            .decode("utf-8")
            .strip()
        )
    except FileNotFoundError:
        logger.error("Git not found or the provided working directory doesn't exist.")
        commit_hash = "N/A"
    except subprocess.CalledProcessError:
        logger.error("The provided directory is not a Git repository.")
        commit_hash = "N/A"

    return commit_hash


def commit_changes(
    commit_message: str,
    file_paths: Optional[List[str]] = None,
    working_dir: Optional[str] = None,
) -> None:
    """
    Commit changes to a Git repository.

    Parameters
    ----------
    commit_message : str
        The message to use for the commit.
    file_paths : Optional[List[str]], default=None
        List of file paths to be added to the commit. If not specified,
        all changes will be committed.
    working_dir : Optional[str], default=None
        The path of the working directory where the Git commands should be executed.
        If None, the commands will be executed in the current working directory.
    """
    log_message_if_working_dir_is_none(working_dir)

    try:
        # Add files to the staging area
        if file_paths is None:
            subprocess.run(["git", "add", "."], check=True, cwd=working_dir)
        else:
            for file_path in file_paths:
                subprocess.run(["git", "add", file_path], check=True, cwd=working_dir)

        # Commit the changes
        subprocess.run(
            ["git", "commit", "-m", commit_message], check=True, cwd=working_dir
        )
    except subprocess.CalledProcessError as error:
        print(f"An error occurred while committing changes: {error}")


def push_changes(
    remote_name: str = "origin",
    branch_name: str = "master",
    working_dir: Optional[str] = None,
) -> None:
    """
    Push committed changes to a remote repository.

    Parameters
    ----------
    remote_name : str, default='origin'
        The name of the remote repository to push to.
    branch_name : str, default='master'
        The name of the branch to push.
    working_dir : str, optional
        The path of the working directory where the Git command should be executed,
        by default None. If None, it uses the current working directory.
    """

    log_message_if_working_dir_is_none(working_dir)

    try:
        # Push the changes
        subprocess.run(
            ["git", "push", remote_name, branch_name], check=True, cwd=working_dir
        )
    except subprocess.CalledProcessError as error:
        print(f"An error occurred while pushing changes: {error}")



"""Decorator Functions."""
import functools
import os
import threading
import time
from datetime import datetime
from typing import Any, Callable, Dict, TypeVar

import numpy as np
import psutil
from fastapi import Request
from prettytable import PrettyTable
from rich.pretty import pprint

#  callable that takes any number of arguments and returns any value.
F = TypeVar("F", bound=Callable[..., Any])


def construct_response(func: F) -> F:
    """Construct a JSON response for an endpoint.

    Supported Frameworks:
    - FastAPI
    To support Flask and Django.

    Reference:
    https://madewithml.com/courses/mlops/api/#decorators
    """

    @functools.wraps(func)
    def wrap(request: Request, *args: Any, **kwargs: Dict[str, Any]) -> Dict[str, Any]:
        results = func(request, *args, **kwargs)
        response = {
            "message": results["message"],
            "method": request.method,
            "status-code": results["status-code"],
            "timestamp": datetime.now().isoformat(),
            "url": request.url,  # ._url
        }
        if "data" in results:
            response["data"] = results["data"]
        return response

    return wrap


# TODO: For memory usage, consider checking out memory_profiler.
# Coding it my own way is not good as it does not take into account
# a lot of minute details, and it does not work for multithreading and
# multiprocessing.
def record_memory_usage(func: F) -> F:
    """Memory usage decorator."""

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Dict[str, Any]) -> Any:
        # Get initial memory usage
        process = psutil.Process(os.getpid())
        mem_info_before = process.memory_info()
        initial_memory = mem_info_before.rss

        result = func(*args, **kwargs)

        # Get final memory usage
        mem_info_after = process.memory_info()
        final_memory = mem_info_after.rss

        memory_used = final_memory - initial_memory

        table = PrettyTable()
        table.field_names = ["Function Name", "Bytes", "Megabytes", "Gigabytes"]
        table.add_row(
            [
                func.__name__,
                f"{memory_used}",
                f"{memory_used / 1024 / 1024}",
                f"{memory_used / 1024 / 1024 / 1024}",
            ]
        )
        pprint(table)

        return result

    return wrapper


class MemoryMonitor:
    def __init__(self, interval=1):
        self.interval = interval  # Time interval in seconds between each check
        self.keep_monitoring = True

    def monitor_memory(self):
        process = psutil.Process(os.getpid())
        while self.keep_monitoring:
            mem_info = process.memory_info()
            print(f"Current memory usage: {mem_info.rss / 1024 / 1024} MB")
            time.sleep(self.interval)

    def start(self):
        self.thread = threading.Thread(target=self.monitor_memory)
        self.keep_monitoring = True
        self.thread.start()

    def stop(self):
        self.keep_monitoring = False
        self.thread.join()


def monitor_memory_usage(func):
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Dict[str, Any]) -> Any:
        monitor = MemoryMonitor()
        monitor.start()

        result = func(*args, **kwargs)

        monitor.stop()

        return result

    return wrapper


@record_memory_usage
@monitor_memory_usage
def increase_memory_usage():
    data = []
    for _ in range(100000):
        data.append("x" * 1000000)  # Increase memory usage by 1 MB each iteration
        # time.sleep(0.1)  # Sleep for a bit to slow down the loop


if __name__ == "__main__":
    increase_memory_usage()



"""
common_utils/core/common.py

This module contains common utility functions for various purposes.
"""
import json
import os
import random
import uuid
from pathlib import Path
from typing import Any, Dict, Optional, Union

import numpy as np
import rich
import torch
import yaml
from dotenv import load_dotenv
from yaml import FullLoader

from common_utils.core.base import DictPersistence
from common_utils.core.logger import Logger

# Setup logging
# logging.basicConfig(
#     level="INFO",
#     format="%(asctime)s [%(levelname)s]: %(message)s",
#     datefmt="%Y-%m-%d %H:%M:%S",
#     handlers=[RichHandler()],
# )

# logger = logging.getLogger("rich")


# Setup logging
logger = Logger(
    module_name=__name__, propagate=False, log_root_dir=None, log_file=None
).logger


class JsonAdapter(DictPersistence):
    def save_as_dict(
        self, data: Dict[str, Any], filepath: str, **kwargs: Dict[str, Any]
    ) -> None:
        """
        Save a dictionary to a specific location.

        Parameters
        ----------
        data : Dict[str, Any]
            Data to save.
        filepath : str
            Location of where to save the data.
        cls : Type, optional
            Encoder to use on dict data, by default None.
        sortkeys : bool, optional
            Whether to sort keys alphabetically, by default False.
        """
        with open(filepath, "w", encoding="utf-8") as f:
            json.dump(data, f, **kwargs)

    def load_to_dict(self, filepath: str, **kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """Load a dictionary from a JSON's filepath.

        Parameters
        ----------
        filepath : str
            Location of the JSON file.

        Returns
        -------
        data: Dict[str, Any]
            Dictionary loaded from the JSON file.
        """
        with open(filepath, "r", encoding="utf-8") as f:
            data = json.load(f, **kwargs)
        return data


class YamlAdapter(DictPersistence):
    def save_as_dict(
        self, data: Dict[str, Any], filepath: str, **kwargs: Dict[str, Any]
    ) -> None:
        with open(filepath, "w", encoding="utf-8") as f:
            yaml.dump(data, f, **kwargs)

    def load_to_dict(self, filepath: str, **kwargs: Dict[str, Any]) -> Dict[str, Any]:
        with open(filepath, "r", encoding="utf-8") as f:
            data = yaml.load(f, Loader=FullLoader, **kwargs)
        return data


def generate_uuid() -> str:
    """Generate a UUID.

    Returns
    -------
    str
        UUID1 as a string.
    """
    return str(uuid.uuid1())


def seed_all(seed: Optional[int] = 1992, seed_torch: bool = True) -> None:
    """
    Seed all random number generators.

    Parameters
    ----------
    seed : int, optional
        Seed number to be used, by default 1992.
    seed_torch : bool, optional
        Whether to seed PyTorch or not, by default True.
    """
    rich.print(f"Using Seed Number {seed}")

    # fmt: off
    os.environ["PYTHONHASHSEED"] = str(seed)        # set PYTHONHASHSEED env var at fixed value
    np.random.seed(seed)                            # numpy pseudo-random generator
    random.seed(seed)                               # python's built-in pseudo-random generator

    if seed_torch:
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.cuda.manual_seed(seed)                # pytorch (both CPU and CUDA)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.enabled = False
    # fmt: on


def seed_worker(_worker_id: int, seed_torch: bool = True) -> None:
    """
    Seed a worker with the given ID.

    Parameters
    ----------
    _worker_id : int
        Worker ID to be used for seeding.
    seed_torch : bool, optional
        Whether to seed PyTorch or not, by default True.

    """
    worker_seed = (
        torch.initial_seed() % 2**32 if seed_torch else random.randint(0, 2**32 - 1)
    )
    np.random.seed(worker_seed)
    random.seed(worker_seed)


def load_env_vars(root_dir: Union[str, Path]) -> Dict[str, str]:
    """
    Load environment variables from .env.default and .env files.

    Parameters
    ----------
    root_dir: Union[str, Path]
        Root directory of the .env files.

    Returns
    -------
    Dict[str, str]
        Dictionary with the environment variables.
    """

    if isinstance(root_dir, str):
        root_dir = Path(root_dir)

    load_dotenv(dotenv_path=root_dir / ".env.default")
    load_dotenv(dotenv_path=root_dir / ".env", override=True)

    return dict(os.environ)


def get_root_dir(env_var: str = "ROOT_DIR", root_dir: str = ".") -> Path:
    """
    Get the root directory of the project.

    Parameters
    ----------
    env_var: str
        Name of the environment variable to use. Defaults to "ROOT_DIR".
    root_dir: str
        Default value to use if the environment variable is not set. Defaults to ".".

    Returns
    -------
    Path
        Path to the root directory of the project.
    """
    root_dir_env = os.getenv(env_var)
    if root_dir_env is None:
        logger.warning("Environment variable %s is not set.", env_var)
        logger.warning("Using default value %s", root_dir)
        return Path(root_dir)
    logger.info(
        "Using environment variable %s and discarding default value %s",
        env_var,
        root_dir,
    )
    return Path(root_dir_env)



"""This module aims to mimic the functionality of Great Expectations, but with a
very minimalistic approach just to illustrate the concept.

TODO: https://cloud.google.com/architecture/mlops-continuous-delivery-and-automation-pipelines-in-machine-learning#data_and_model_validation
    - Add more validation methods.
"""
from __future__ import annotations

import textwrap
from typing import Any, Dict

import pandas as pd
from tabulate import tabulate

from common_utils.core.logger import Logger

# Setup logging
logger = Logger(
    module_name=__name__, propagate=False, log_root_dir=None, log_file=None
).logger


class DataFrameValidator:
    """Validate pandas DataFrame against specified schema.

    Attributes
    ----------
    df : pd.DataFrame
        The dataframe to be validated.
    schema : Dict[str, Any]
        The expected schema of the dataframe as a dictionary.
        The dictionary keys should be column names and the values should be the expected dtypes.
    """

    def __init__(self, df: pd.DataFrame, schema: Dict[str, Any]) -> None:
        self.df = df
        self.schema = schema

    def validate_missing(self) -> DataFrameValidator:
        """Check if there are missing values in the dataframe."""
        logger.info("Checking missing values...")
        # note missing is the number of missing values per column
        missing: pd.Series = self.df.isnull().sum().sort_values(ascending=False)

        percent: pd.Series = (
            self.df.isnull().sum() / self.df.isnull().count()
        ).sort_values(ascending=False)

        missing_data = pd.concat(
            [missing, percent], axis=1, keys=["Missing", "Percent"]
        )
        missing_data.index.name = "Feature"

        if missing_data["Missing"].any():
            logger.warning(  # pylint: disable=logging-not-lazy
                "The following columns have missing data:\n"
                + textwrap.indent(
                    tabulate(
                        missing_data[missing_data["Missing"] > 0],
                        headers="keys",
                        tablefmt="psql",
                    ),
                    "    ",
                )
            )
        return self

    def validate_data_types(self) -> DataFrameValidator:
        """Check if the data types of the dataframe's columns match the expected schema."""
        logger.info("Checking data types...")
        for column, dtype in self.schema.items():
            if self.df[column].dtype != dtype:
                logger.warning(
                    f"The data type for {column} is not {dtype}, it's {self.df[column].dtype}"
                )
        return self

    def validate_schema(self) -> DataFrameValidator:
        """Check if the schema of the dataframe matches the expected schema."""
        logger.info("Checking schema...")
        for column in self.schema.keys():
            if column not in self.df.columns:
                logger.warning(
                    f"Expected {column} in the dataframe, but it's not present."
                )
        for column in self.df.columns:
            if column not in self.schema.keys():
                logger.warning(f"Found unexpected column {column} in the dataframe.")
        return self

    def validate_all(self) -> None:
        """Perform all validations on the dataframe."""
        self.validate_schema().validate_data_types().validate_missing()



"""See https://cloud.google.com/architecture/mlops-continuous-delivery-and-automation-pipelines-in-machine-learning
model validation section.

This script handles promoting a model to production.
Simple logic is if the new model is better than the current production model, then promote it to production.
Of course in real life, you might want to implement more robust model comparison logic and
implement a manual approval step before promoting the model to production. Include A/B testing too.
"""

from typing import List, Literal, Optional

from mlflow.tracking import MlflowClient

from common_utils.core.logger import Logger
from common_utils.experiment_tracking.promoter.base import (
    ModelVersion,
    PromotionManager,
)


class MLFlowPromotionManager(PromotionManager):
    """A class that manages the promotion of machine learning models to production.

    This class adheres to several principles of good software design:

    1. Single Responsibility Principle (SRP): This class's responsibility is clear -
       managing the promotion of models to production. It maintains the current and
       previous versions of a model, and decides whether to promote a new model
       based on a comparison of their performances.

    2. Open-Closed Principle (OCP): This class is open for extension (e.g., one can
       easily extend it to use different promotion strategies) but closed for
       modification (adding new functionality doesn't require modification of the
       existing code).

    3. Liskov Substitution Principle (LSP): This principle is not explicitly
       implemented in this class since it doesn't have any subclasses. However, if a
       subclass were created (e.g., a `ThresholdPromotionManager` that promotes
       models that exceed a certain performance threshold), it could be substituted
       for this class without altering any of the properties that this class
       guarantees (e.g., that a model won't be demoted).

    4. Interface Segregation Principle (ISP): The `ModelPromotionManager` class does
       not depend on any interfaces it doesn't use. The class itself can be seen as
       a high-level interface for model promotion operations.

    5. Dependency Inversion Principle (DIP): The `ModelPromotionManager` class
       depends on abstractions (e.g., the `MlflowClient` class and the
       `ModelVersion` class), not on concrete classes or implementations.

    NOTE: For design pattern consider using the Strategy Pattern.
    To allow for various types of clients like Mlflow, Wandb, etc., you could employ
    the Strategy Design Pattern. This pattern enables a strategy (algorithm) to be
    selected at runtime. It is one of the behavioral design patterns and it's quite
    useful when you have a set of similar algorithms and want to switch between them
    dynamically.

    In the context of your ModelPromotionManager class, the "strategy" would be the
    different types of clients you want to support (like Mlflow, Wandb, etc.). You
    could define an interface that each client must implement. Then, in your
    ModelPromotionManager class, you would use this interface instead of a specific
    client implementation.

    This change would make your ModelPromotionManager class adhere more closely to
    the Open/Closed Principle (part of the SOLID principles), because it would be
    open for extension (you can add support for new types of clients) but closed for
    modification (you don't have to change the ModelPromotionManager class to add
    new clients).
    """

    def __init__(
        self,
        client: MlflowClient,
        model_name: str,
        logger: Optional[Logger] = None,
    ) -> None:
        self.client = client
        self.model_name = model_name

        if logger is None:
            self.logger = Logger(
                module_name=__name__, propagate=False, log_root_dir=None, log_file=None
            ).logger
        else:
            self.logger = logger

    def _check_if_there_exists_production_model(self) -> Literal[True, False]:
        return len(self._get_all_production_models()) > 0

    def _get_all_production_models(self) -> List:
        production_models = self.client.get_latest_versions(
            self.model_name, stages=["Production"]
        )
        return production_models

    def _get_latest_production_model(self) -> ModelVersion:
        """
        Fetch the current production model, if one exists.
        Assume latest production model is the one with the highest version number
        and also the one deployed most recently.
        """
        production_models = self._get_all_production_models()
        sorted_production_models = sorted(
            production_models, key=lambda model: model.version, reverse=True
        )

        # get the latest production model
        latest_production_model = sorted_production_models[0]
        run_id = latest_production_model.run_id
        run = self.client.get_run(run_id)
        return ModelVersion(
            version=latest_production_model.version,
            metrics=run.data.metrics,
            stage=latest_production_model.current_stage,
        )

    def _find_best_model_for_production(
        self, metric_name: str = "test_f1"
    ) -> ModelVersion:
        """
        Find the best model to promote to production based on the metric score.
        This is needed for the first time when there is no production models.
        """
        model_versions = self.client.search_model_versions(f"name='{self.model_name}'")
        model_versions_metrics = [
            (
                version.version,
                self.client.get_run(version.run_id).data.metrics.get(metric_name),
            )
            for version in model_versions
        ]
        model_versions_metrics = [x for x in model_versions_metrics if x[1] is not None]
        best_model_version, best_model_metrics = max(
            model_versions_metrics, key=lambda x: x[1]
        )
        return ModelVersion(best_model_version, best_model_metrics, "Production")

    def _compare_models(
        self,
        model1: ModelVersion,
        model2: ModelVersion,
        metric_name: str = "test_f1",
    ) -> bool:
        """
        Compare two models based on their performance metric.

        Args:
            model1: The first model to compare.
            model2: The second model to compare.
            metric_name: The name of the performance metric to use for the comparison.

        Returns:
            bool: True if model1's performance is better than model2's performance, False otherwise.
        """
        return model1.metrics.get(metric_name, 0) > model2.metrics.get(metric_name, 0)

    def _transition_model_to_production(self, model_version: int) -> None:
        """
        Transition a model to the production stage.

        Args:
            model_version: The version number of the model to transition to production.
        """
        self.client.transition_model_version_stage(
            name=self.model_name,
            version=model_version,
            stage="Production",
        )
        self.logger.info(
            f"Model {self.model_name} version {model_version} is now in production."
        )

    def promote_to_production(self, metric_name: str = "test_f1") -> None:
        """
        Promote the best model version to production if it outperforms the current production model.

        Args:
            metric_name: The name of the performance metric to use for the comparison.
        """
        has_production_model: bool = self._check_if_there_exists_production_model()

        # no production model yet, promote the best model in the current stage
        if not has_production_model:
            # Find the model version with the highest test_f1 score
            curr_best_model: ModelVersion = self._find_best_model_for_production(
                metric_name=metric_name
            )
            self._transition_model_to_production(curr_best_model.version)
            # exit
            return
        else:
            latest_production_model = self._get_latest_production_model()

            # get current model, note we do not need to get all the model versions
            # because we assume the latest model version is the one we want to promote
            non_production_latest_model = self.client.get_latest_versions(
                self.model_name, stages=["None"]
            )[0]

            # get the model version number and metrics
            latest_curr_model_version = non_production_latest_model.version
            latest_curr_model_run_id = non_production_latest_model.run_id
            latest_curr_model_stage = non_production_latest_model.current_stage
            latest_curr_model_run = self.client.get_run(latest_curr_model_run_id)

            latest_curr_model_metrics = latest_curr_model_run.data.metrics
            latest_curr_model: ModelVersion = ModelVersion(
                version=latest_curr_model_version,
                metrics=latest_curr_model_metrics,
                stage=latest_curr_model_stage,
            )

            if not self._compare_models(
                model1=latest_curr_model, model2=latest_production_model
            ):
                self.logger.info(
                    f"Model {self.model_name} version {latest_curr_model_version} does not outperform the current production model."
                )
                return
            self._transition_model_to_production(latest_curr_model_version)



from abc import ABC, abstractmethod
from typing import Any, Dict


class DictPersistence(ABC):
    """Abstract class for saving and loading dictionary."""

    @abstractmethod
    def save_as_dict(
        self, data: Dict[str, Any], filepath: str, **kwargs: Dict[str, Any]
    ) -> None:
        """Save a dictionary to a specific location."""

    @abstractmethod
    def load_to_dict(self, filepath: str, **kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """Load a dictionary from a filepath."""


class Connection(ABC):
    """Abstract class for database connection.

    See common_utils/cloud/gcp/database/bigquery.py for example.

    NOTE:
        1. In particular, it should have a property or attribute table_name.
    """

    @abstractmethod
    def connect(self) -> None:
        """Connect to a database."""

    @abstractmethod
    def disconnect(self) -> None:
        """Disconnect from a database."""

    @abstractmethod
    def query(self, query: str) -> Any:
        """Query a database."""


class Storage(ABC):
    """Abstract class for storage.

    See common_utils/cloud/gcp/storage/gcs.py for example.

    NOTE:
        1. Subsequent cloud storages should abide by this interface. This means
            turning S3, Azure, etc. into a class that inherits from this class.
    """

    @abstractmethod
    def upload_blob(self) -> None:
        """Save a file to a specific location."""

    @abstractmethod
    def upload_blobs(self) -> None:
        """Save multiple files to a specific location."""

    @abstractmethod
    def upload_directory(self) -> None:
        """Save a directory to a specific location."""

    @abstractmethod
    def download_blob(self) -> None:
        """Load a file from a specific location."""

    @abstractmethod
    def download_blobs(self) -> None:
        """Load multiple files from a specific location."""



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Literal, Optional


@dataclass(frozen=True)
class ModelVersion:
    """
    A class representing a version of an ML model.

    This class follows the Value Object pattern, as each instance represents
    a version of a model and its associated properties without a specific
    identity.

    NOTE: Be aware of potential namespace conflicts with Mlflow's
    ModelVersion class (ModelVersion).

    Single Responsibility Principle is followed, as the class solely
    encapsulates the logic related to a model version.

    Attributes
    ----------
    version : int
        Version number, usually incremental.
    metrics : Dict[str, float]
        Performance metrics as a dictionary, e.g., {'accuracy': 0.95}.
    stage : str
        Stage of the model version, e.g., 'production'.
    """

    version: int
    metrics: Dict[str, float]
    stage: Optional[str] = None


class PromotionManager(ABC):
    """
    A base class (interface) for a model promotion manager.

    This class outlines the basic methods any concrete promotion manager should implement.
    """

    @abstractmethod
    def _check_if_there_exists_production_model(self) -> Literal[True, False]:
        ...

    @abstractmethod
    def _get_all_production_models(self) -> List[ModelVersion]:
        ...

    @abstractmethod
    def _get_latest_production_model(self) -> ModelVersion:
        ...

    @abstractmethod
    def _find_best_model_for_production(self, metric_name: str) -> ModelVersion:
        ...

    @abstractmethod
    def _compare_models(
        self,
        model1: ModelVersion,
        model2: ModelVersion,
        metric_name: str,
    ) -> Literal[True, False]:
        ...

    @abstractmethod
    def _transition_model_to_production(self, model_version: int) -> None:
        ...

    @abstractmethod
    def promote_to_production(self, metric_name: str) -> None:
        """The main method to promote a model to production."""



"""Decorator Functions."""
import functools
import time
from statistics import mean, median, stdev
from typing import Any, Callable, Dict, TypeVar

import numpy as np
from prettytable import PrettyTable
from rich.pretty import pprint

from common_utils.core.logger import Logger

# Setup logging
LOGGER = Logger(
    module_name=__name__, propagate=False, log_root_dir=None, log_file=None
).logger


F = TypeVar("F", bound=Callable[..., Any])


# probably the only time when you don't use CamelCase for class names
class timer:
    def __init__(
        self,
        display_table: bool = True,
        unit: str = "seconds",
        decimal_places: int = 4,
        store_times: bool = False,
        log: bool = False,
    ) -> None:
        self.display_table = display_table
        self.unit = unit
        self.decimal_places = decimal_places
        self.store_times = store_times
        self.execution_times = {}
        self.log = log

    def __call__(self, func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Dict[str, Any]) -> Any:
            start_time = time.time()
            result = func(*args, **kwargs)
            end_time = time.time()
            elapsed_time = end_time - start_time
            elapsed_time = convert_time_unit(elapsed_time, self.unit)

            if self.store_times:
                if func.__name__ not in self.execution_times:
                    self.execution_times[func.__name__] = []
                self.execution_times[func.__name__].append(elapsed_time)

            if self.display_table:
                # Create a table to display the results
                table = PrettyTable()
                table.field_names = [
                    "Function Name",
                    f"Elapsed Time ({self.unit.capitalize()})",
                ]
                table.add_row(
                    [
                        func.__name__,
                        f"{round(elapsed_time, self.decimal_places):.{self.decimal_places}f}",
                    ]
                )

                if self.store_times and len(self.execution_times[func.__name__]) > 1:
                    times = self.execution_times[func.__name__]
                    stats_table = PrettyTable()
                    stats_table.field_names = ["Mean", "Median", "Stdev"]
                    stats_table.add_row(
                        [
                            round(mean(times), self.decimal_places),
                            round(median(times), self.decimal_places),
                            round(stdev(times), self.decimal_places),
                        ]
                    )
                    table.add_row(["Statistics", stats_table.get_string()])

                pprint(table)

            # Log the results
            if self.log:
                LOGGER.info(
                    f"Function {func.__name__} executed in "
                    f"{elapsed_time} seconds with args: {args} and "
                    f"kwargs: {kwargs}"
                )
            return result

        return wrapper


def convert_time_unit(time_in_seconds: float, unit: str) -> float:
    """Converts time to the desired unit."""
    if unit == "minutes":
        return time_in_seconds / 60
    if unit == "hours":
        return time_in_seconds / 60 / 60
    if unit == "seconds":
        return time_in_seconds

    raise ValueError("Unknown time unit. Please use 'seconds', 'minutes', or 'hours'.")


@timer(
    display_table=True, unit="seconds", decimal_places=4, store_times=True, log=False
)
def add_two_arrays(array_1: np.ndarray, array_2: np.ndarray) -> np.ndarray:
    """Add two arrays together."""
    return array_1 + array_2


if __name__ == "__main__":
    array_1 = np.random.randint(0, 100, size=(10000, 10000))
    array_2 = np.random.randint(0, 100, size=(10000, 10000))
    REPEAT = 1
    for _ in range(REPEAT):
        add_two_arrays(array_1, array_2)



# core/directory_utils.py
from pathlib import Path
from typing import Union

import pandas as pd
import rich
from rich.pretty import pprint


def get_size(directory: Path) -> int:
    """Get the total size of a directory including all its subdirectories and files."""
    return sum(f.stat().st_size for f in directory.glob("**/*") if f.is_file())


def create_directory_size_table(root_directory: str) -> pd.DataFrame:
    """Create a pandas DataFrame with the name and size of each subdirectory in a root directory."""
    root = Path(root_directory)
    data = {"Name": [], "Size (bytes)": [], "Size (MB)": [], "Size (GB)": []}
    size_units = [
        1,
        1024**2,
        1024**3,
    ]  # factors to convert bytes to bytes, MB, and GB

    for directory in root.iterdir():
        if directory.is_dir():
            size_bytes = get_size(directory)
            data["Name"].append(directory.name)
            for size_type, unit in zip(
                ["Size (bytes)", "Size (MB)", "Size (GB)"], size_units
            ):
                data[size_type].append(size_bytes / unit)

    # Calculate total sizes
    total_sizes = {
        size_type: sum(sizes)
        for size_type, sizes in data.items()
        if size_type != "Name"
    }
    data["Name"].append("Total")
    for size_type, total_size in total_sizes.items():
        data[size_type].append(total_size)

    df = pd.DataFrame(data)
    pprint(df)
    return df


def list_files_recursively(start_path: Union[str, Path]) -> None:
    """
    List all files and directories recursively in the given path using markdown
    style.

    Parameters
    ----------
    start_path : Union[str, Path]
        The path where the function should start listing the files and
        directories.

    Returns
    -------
    None
    """

    start_path = Path(start_path)

    def _list_files(path: Path, level: int, is_last: bool) -> None:
        """
        Helper function to list files and directories at the given path.

        Parameters
        ----------
        path : Path
            The path to list files and directories from.
        level : int
            The current depth in the file hierarchy.
        is_last : bool
            Indicates whether the current path is the last item in its parent
            directory.

        Returns
        -------
        None
        """
        prefix = (
            "    " * (level - 1) + ("└── " if is_last else "├── ") if level > 0 else ""
        )
        rich.print(f"{prefix}{path.name}/")
        children = sorted(list(path.iterdir()), key=lambda x: x.name)
        for i, child in enumerate(children):
            if child.is_file():
                child_prefix = "    " * level + (
                    "└── " if i == len(children) - 1 else "├── "
                )
                rich.print(f"{child_prefix}{child.name}")
            elif child.is_dir():
                _list_files(child, level + 1, i == len(children) - 1)

    _list_files(start_path, 0, False)



from dataclasses import dataclass, field

from google.auth.credentials import Credentials
from google.oauth2 import service_account


# pylint: disable=too-few-public-methods
@dataclass
class GCPConnector:
    """
    A class to handle connections and operations on Google Cloud Platform
    (BigQuery and GCS).

    Attributes
    ----------
    project_id : str
        The project ID associated with the GCP services.
    google_application_credentials : str
        The path to the service account key JSON file.
    """

    # ellipsis (...) means required
    project_id: str
    google_application_credentials: str
    credentials: Credentials = field(init=False)

    def __post_init__(self) -> None:
        """
        Initialize the credentials for the GCP services.
        """
        self.credentials = service_account.Credentials.from_service_account_file(
            self.google_application_credentials
        )



from __future__ import annotations
import datetime
import time
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import Callable, Any, Dict
import matplotlib.pyplot as plt
import networkx as nx


# TODO: how to do parallelism?
class Task:
    """This is a Node in the graph."""

    def __init__(self, func: Callable, pipeline: Pipeline) -> None:
        self.func = func
        self.pipeline = pipeline

        self.name: str = func.__name__

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.func(*args, **kwargs)

    def __rshift__(self, other: Task) -> Task:
        self.pipeline.add_dependency(self, other)
        return other


class Pipeline:
    def __init__(self):
        self.graph = nx.DiGraph()
        self.tasks = {}
        self.task_data = {}

    def task(self):
        def decorator(f):
            task = Task(f, self)
            self.tasks[task.name] = task
            self.graph.add_node(task.name)
            return task

        return decorator

    def add_dependency(self, upstream, downstream):
        self.graph.add_edge(upstream.name, downstream.name)

    def run(self):
        with ThreadPoolExecutor(max_workers=3) as executor:
            for task_name in nx.topological_sort(self.graph):
                task_func = self.tasks[task_name]
                if task_name in self.task_data:
                    args = self.task_data[task_name]
                else:
                    args = ()
                future = executor.submit(task_func, *args)
                result = future.result()
                for successor in self.graph.successors(task_name):
                    self.task_data[successor] = (result,)
            return self.task_data

    def visualize(self) -> None:
        pos = nx.spring_layout(self.graph)
        for node in self.graph.nodes:
            plt.text(
                pos[node][0],
                pos[node][1],
                node,
                fontsize=12,
                bbox=dict(
                    facecolor="blue", edgecolor="black", boxstyle="round,pad=0.2"
                ),
            )
        nx.draw(self.graph, pos, with_labels=False, node_color="w", edgecolors="k")
        plt.show()


pipeline = Pipeline()


@pipeline.task()
def feature_pipeline():
    print(f"{datetime.datetime.now()} - Running feature pipeline")
    time.sleep(2)
    print(f"{datetime.datetime.now()} - Finished feature pipeline")
    return "features"


@pipeline.task()
def training_pipeline(features):
    print(f"{datetime.datetime.now()} - Running training pipeline with {features}")
    time.sleep(2)
    print(f"{datetime.datetime.now()} - Finished training pipeline")
    return "trained_model"


@pipeline.task()
def prediction_pipeline(trained_model):
    print(
        f"{datetime.datetime.now()} - Running prediction pipeline with {trained_model}"
    )
    time.sleep(5)
    print(f"{datetime.datetime.now()} - Finished prediction pipeline")
    return "predictions"


@pipeline.task()
def evaluation_pipeline(trained_model):
    print(
        f"{datetime.datetime.now()} - Running evaluation pipeline with {trained_model}"
    )
    time.sleep(5)
    print(f"{datetime.datetime.now()} - Finished evaluation pipeline")
    return "evaluation_metrics"


@pipeline.task()
def validation_pipeline(trained_model):
    print(
        f"{datetime.datetime.now()} - Running validation pipeline with {trained_model}"
    )
    time.sleep(5)
    print(f"{datetime.datetime.now()} - Finished validation pipeline")
    return "validation_metrics"


feature_pipeline >> training_pipeline >> prediction_pipeline
feature_pipeline >> training_pipeline >> evaluation_pipeline
feature_pipeline >> training_pipeline >> validation_pipeline


# pipeline.visualize()
pipeline.run()



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Literal

from google.cloud import storage
from google.cloud.exceptions import NotFound

from common_utils.cloud.base import GCPConnector
from common_utils.core.logger import Logger

# Setup logging
# logging.basicConfig(
#     level="INFO",
#     format="%(asctime)s [%(levelname)s]: %(message)s",
#     datefmt="%Y-%m-%d %H:%M:%S",
#     handlers=[RichHandler()],
# )

# logger = logging.getLogger("rich")


# Setup logging
logger = Logger(
    module_name=__name__, propagate=False, log_root_dir=None, log_file=None
).logger


@dataclass
class GCS(GCPConnector):
    bucket_name: str
    storage_client: storage.Client = field(init=False, repr=False)
    bucket: storage.Bucket = field(init=False, repr=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.storage_client = storage.Client(
            credentials=self.credentials, project=self.project_id
        )
        self._init_bucket(self.bucket_name)

    def _init_bucket(self, bucket_name: str) -> None:
        """
        Initialize a GCS bucket.

        Parameters
        ----------
        bucket_name : str
            The name of the GCS bucket.

        Returns
        -------
        None
        """
        self.bucket = self.storage_client.bucket(bucket_name)

    def check_if_bucket_exists(self) -> Literal[True, False]:
        """Check if a bucket exists."""
        try:
            self.storage_client.get_bucket(self.bucket_name)
            logger.info(f"Bucket {self.bucket_name} already exists")
            return True
        except NotFound:
            logger.info(
                f"""Bucket {self.bucket_name} does not exist.
                Please create it using `create_bucket`."""
            )
            return False

    def create_bucket(self) -> None:
        """Creates a new GCS bucket if it doesn't exist."""
        self.storage_client.create_bucket(self.bucket_name)
        logger.info(f"Bucket {self.bucket_name} created")

    def list_gcs_files(self, prefix: str = "", **kwargs: Dict[str, Any]) -> List[str]:
        """
        List the files in a GCS bucket with an optional prefix.

        Parameters
        ----------
        bucket_name : str
            The name of the GCS bucket.
        prefix : str, optional, default: ""
            The prefix to filter files in the bucket, by default "".
        **kwargs : Dict[str, Any]
            Additional arguments to pass to the list_blobs method.

        Returns
        -------
        gcs_files: List[str]
            The list of file names in the specified GCS bucket.
        """
        blobs = self.storage_client.list_blobs(
            self.bucket_name, prefix=prefix, **kwargs
        )
        gcs_files = [blob.name for blob in blobs]
        return gcs_files

    def create_blob(self, destination_blob_name: str) -> storage.Blob:
        """Creates a new blob in the bucket. See upload_blob for example."""
        blob = self.bucket.blob(destination_blob_name)
        return blob

    def upload_blob(
        self,
        source_file_name: str,
        destination_blob_name: str,
        **kwargs: Dict[str, Any],
    ) -> None:
        """
        Uploads a file to a GCS bucket.
        https://cloud.google.com/storage/docs/uploading-objects#storage-upload-object-client-libraries

        # The ID of your GCS bucket
        # bucket_name = "your-bucket-name"
        # The path to your file to upload
        # source_file_name = "local/path/to/file"
        # The ID of your GCS object
        # destination_blob_name = "storage-object-name"

        Parameters
        ----------
        bucket_name : str
            The ID of your GCS bucket.
        source_file_name : str
            The path to your file to upload.
        destination_blob_name : str
            The ID of your GCS object.

        Returns
        -------
        None
        """
        # Optional: set a generation-match precondition to avoid potential race conditions
        # and data corruptions. The request to upload is aborted if the object's
        # generation number does not match your precondition. For a destination
        # object that does not yet exist, set the if_generation_match precondition to 0.
        # If the destination object already exists in your bucket, set instead a
        # generation-match precondition using its generation number.
        generation_match_precondition = 0

        blob = self.create_blob(destination_blob_name)

        blob.upload_from_filename(
            source_file_name,
            if_generation_match=generation_match_precondition,
            **kwargs,
        )

    def upload_directory(
        self, source_dir: str, destination_dir: str, **kwargs: Dict[str, Any]
    ) -> None:
        """
        Uploads a directory to a GCS bucket.
        https://cloud.google.com/storage/docs/uploading-objects
        """
        for file_path in Path(source_dir).glob("**/*"):
            if file_path.is_file():
                destination_blob_name = (
                    destination_dir + "/" + str(file_path.relative_to(source_dir))
                )
                self.upload_blob(str(file_path), destination_blob_name, **kwargs)

    def download_blob(
        self,
        source_blob_name: str,
        destination_file_name: str,
        **kwargs: Dict[str, Any],
    ) -> None:
        """
        Downloads a blob from the bucket.
        https://cloud.google.com/storage/docs/downloading-objects

        # The ID of your GCS bucket
        # bucket_name = "your-bucket-name"

        # The ID of your GCS object
        # source_blob_name = "storage-object-name"

        # The path to which the file should be downloaded
        # destination_file_name = "local/path/to/file"
        """

        # Construct a client side representation of a blob.
        # Note `Bucket.blob` differs from `Bucket.get_blob` as it doesn't retrieve
        # any content from Google Cloud Storage. As we don't need additional data,
        # using `Bucket.blob` is preferred here.
        blob = self.create_blob(source_blob_name)

        blob.download_to_filename(destination_file_name, **kwargs)



from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import pandas as pd
from google.cloud import bigquery
from google.cloud.exceptions import NotFound

from common_utils.cloud.base import GCPConnector
from common_utils.core.logger import Logger

# Setup logging
logger = Logger(
    module_name=__name__, propagate=False, log_root_dir=None, log_file=None
).logger


@dataclass
class BigQuery(GCPConnector):
    # usually you call project_id.dataset_id.table_name
    dataset: str  # The ID of the dataset to use.
    table_name: str  # The name of the table to use.
    _dataset_id: str = field(init=False)  # The ID of the dataset to use.
    _table_id: str = field(init=False)  # The full ID of the table to use.
    bigquery_client: bigquery.Client = field(init=False)

    def __post_init__(self) -> None:
        """
        Initialize the BigQuery client for the GCP services.
        """
        super().__post_init__()
        self.bigquery_client = bigquery.Client(
            credentials=self.credentials, project=self.project_id
        )

    @property
    def table_id(self) -> str:
        """Return the full ID of the table to use."""
        return f"{self.project_id}.{self.dataset}.{self.table_name}"

    @property
    def dataset_id(self) -> str:
        """Return the ID of the dataset to use."""
        return f"{self.project_id}.{self.dataset}"

    def check_if_dataset_exists(self) -> Literal[True, False]:
        """Check if a dataset exists."""
        try:
            self.bigquery_client.get_dataset(self.dataset_id)
            logger.info(f"Dataset {self.dataset_id} already exists")
            return True
        except NotFound:
            logger.warning(
                f"""Dataset {self.dataset_id} does not exist.
                Please create it using `create_dataset`."""
            )
            return False

    def create_dataset(self) -> None:
        """Creates a new BigQuery dataset if it doesn't exist."""
        dataset = bigquery.Dataset(self.dataset_id)
        self.bigquery_client.create_dataset(dataset)
        logger.info(f"Created dataset {self.dataset_id}")

    def check_if_table_exists(self) -> Literal[True, False]:
        """Check if a table exists."""
        try:
            self.bigquery_client.get_table(self.table_id)
            logger.info(f"Table {self.table_id} already exists")
            return True
        except NotFound:
            logger.warning(
                f"""Table {self.table_id} does not exist.
                Please create it using `create_table`."""
            )
            return False

    def create_table(self, schema: List[bigquery.SchemaField]) -> None:
        """Creates a new BigQuery table if it doesn't exist."""
        table = bigquery.Table(self.table_id, schema=schema)
        self.bigquery_client.create_table(table)
        logger.info(f"Created table {self.table_id}")

    def query(
        self, query: str, as_dataframe: bool = True
    ) -> Union[List[Tuple[Any]], pd.DataFrame]:
        """
        Execute a query in BigQuery and return the result as a DataFrame.

        Parameters
        ----------
        query : str
            The SQL query to execute in BigQuery.

        Returns
        -------
        pd.DataFrame
            The result of the query as a DataFrame.
        """
        query_job = self.bigquery_client.query(query)
        results = query_job.result()
        return results.to_dataframe() if as_dataframe else results

    def load_job_config(
        self,
        schema: Optional[List[bigquery.SchemaField]] = None,
        write_disposition: str = "WRITE_APPEND",
        **kwargs: Dict[str, Any],
    ) -> bigquery.LoadJobConfig:
        """
        Creates a load job configuration for BigQuery.
        """
        return bigquery.LoadJobConfig(
            schema=schema, write_disposition=write_disposition, **kwargs
        )

    def load_table_from_dataframe(
        self,
        df: pd.DataFrame,
        job_config: bigquery.LoadJobConfig,
        **kwargs,
    ) -> None:
        """
        Loads data from a DataFrame to BigQuery.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame to load.
        table_id : str
            The full ID of the table where the data will be loaded.
        schema: List[SchemaField]
            The schema fields to use for the table. None if the table already exists.
        """
        load_job = self.bigquery_client.load_table_from_dataframe(
            df, self.table_id, job_config=job_config, **kwargs
        )

        load_job.result()  # Waits for the job to complete.

        logger.info(f"Loaded {load_job.output_rows} rows into {self.table_id}.")
